﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.CompilerServices;

namespace PuzzlePlayer_Namespace
{ 
    /* The binair board consist of 1 and 0
     * This means that the possible states a cell can be are: empty, one or zero
     * To specify this in an int[,] array we will use -1 for an empty space,
     * 0 for a space with a zero and 1 for a space with a one
     * The empty space is a constant defined in the abstract Board class
     */

    internal class Binary : Board
    {

        // constructor with baordSize parameter (default is set to 8 but can be changed)
        public Binary(int boardSize = 8)
        {
            // create a board with the specifide size
            boardState = new int[boardSize,boardSize];

            description = "Binary puzzle is played on any even-numbered square grid, with some cells initially containing black or white circles. The goal of the puzzle is to fill all cells such that:\r\n- More than two circles of the same color cannot be adjacent\r\n- Each row and column must contain an equal number of black and white circles\r\n- Each row and column cannot appear multiple times on the board";

            // clear the board (fill it in with -1)
            Clear();
        }

        public override void Draw (Graphics gr, Point p, Size s)
        {
            gr.FillRectangle(Brushes.Beige, p.X, p.Y, s.Width, s.Height);
            for (int i = 0; i < boardState.GetLength(0); i++)
            {
                for(int j = 0; j < boardState.GetLength(1); j++)
                {
                    gr.DrawRectangle(Pens.Black,
                        p.X+i*s.Width/boardState.GetLength(0),
                        p.Y+j*s.Height/boardState.GetLength(1),
                        s.Width / boardState.GetLength(0),
                        s.Height / boardState.GetLength(1));
                    if (boardState[i,j] == 0)
                    {
                        gr.FillEllipse(Brushes.White,
                            (int)(p.X + ((double)i + 0.125) * s.Width / boardState.GetLength(0)),
                            (int)(p.Y + ((double)j + 0.125) * s.Height / boardState.GetLength(1)),
                            s.Width / boardState.GetLength(0) * 3 / 4,
                            s.Height / boardState.GetLength(1) * 3 / 4);
                    }
                    if (boardState[i, j] == 1)
                    {
                        gr.FillEllipse(Brushes.Black,
                            (int)(p.X + ((double)i + 0.125) * s.Width / boardState.GetLength(0)),
                            (int)(p.Y + ((double)j + 0.125) * s.Height / boardState.GetLength(1)),
                            s.Width / boardState.GetLength(0) * 3 / 4,
                            s.Height / boardState.GetLength(1) * 3 / 4);
                    }
                }
            }
        }

        private void Clear()
        {
            int size = boardState.GetLength(0);
            // fill the board with empty spaces (-1)
            for (int i = 0; i < size; i++)
            {
                for (int j = 0; j < size; j++)
                    boardState[i, j] = emptySpace;
            }
        }

        /* this static methode returns true if the board is valid in a few steps
        1: It checks if the size from boardToCheck is equal to the size from the boardState variable
        2: it loops though the board and checks if one of the spaces contains something other than a -1,0 or 1
        */
        public override bool IsBoardValid(int[,] boardToCheck)
        {
            // check if the size is NOT the same
            if(!(boardToCheck.GetLength(0) ==  boardState.GetLength(0) && boardToCheck.GetLength(1) == boardState.GetLength(1)))
                return false;
                

            // check if any of the spaces doesn't contain a -1,0 or 1
            for(int i = 0;i < boardToCheck.GetLength(0);i++)
            {
                for(int j = 0; j < boardToCheck.GetLength(1);j++)
                {
                    switch(boardToCheck[i,j])
                    {
                        case -1:
                        case 0:
                        case 1:
                            break;
                        default:
                            return false;
                    }
                }
            }

            return true;
        }

        public override void Generate()
        {
            throw new NotImplementedException();
        }

        protected override List<int[,]> GetSolveList(int[,] boardToSolve)
        {
            List<int[,]> result = new List<int[,]>();

            for (int i = 0; i < boardToSolve.GetLength(0); i++)
                for (int j = 0; j < boardToSolve.GetLength(1); j++)
                {
                    int[,] move = CheckMove(i, j, boardToSolve);
                    if (move != null)
                        result.Add(move);
                }

            return result;
        }

        private int[,] CheckMove(int x, int y, int[,] boardToSolve)
        {
            int[,] result = boardToSolve;

            bool validForZero = false;
            bool validForOne = false;

            // empty check
            if (boardToSolve[x, y] != emptySpace)
                return null;

            // loop two times for checking 0 and 1
            for (int i = 0; i <= 1; i++)
            {
                int opposite;

                if (i == 0)
                    opposite = 1;
                else
                    opposite = 0;

                // check if one of the checks succeded
                bool valid = MiddleCheck(x, y, boardToSolve, opposite) ||
                    SideCheck(x, y, boardToSolve, opposite) ||
                    EvenCheck(x, y, boardToSolve, opposite);

                if (i == 0)
                    validForZero = valid;
                else
                    validForOne = valid;
            }

            // if both 1 and 0 are valid then something went wrong
            if (validForZero && validForOne)
                return null;

            // change the specifided place in a 0 or 1 if they are valid. if both are not valid then there is no valid move on the checked space and null is returnd.
            if (validForZero)
                result[x, y] = 0;
            else if (validForOne)
                result[x, y] = 1;
            else
                return null;

            return result;
        }

        // check if the space is surrounded on both sides by the same number. If it is, the checked space should be the opposite number
        private bool MiddleCheck(int x, int y, int[,] b, int checkFor)
        {
            // first check if x-1 and x+1 aren't out of bounds
            // after that do the check if the move is valid
            if(!(x-1 < 0 || x+1 > b.GetLength(0) - 1))
                if (b[x - 1, y] == checkFor && b[x + 1, y] == checkFor)
                    return true;

            // same for y
            if (!(y - 1 < 0 || y + 1 > b.GetLength(1) - 1))
                if (b[x, y - 1] == checkFor && b[x, y + 1] == checkFor)
                    return true;

            // return false if nothing was found
            return false;
        }

        // check if the two spaces left, right, up or down of the space are the opposite number. if so return true
        private bool SideCheck(int x, int y, int[,] b, int checkFor)
        {
            if (!(x - 2 < 0 || x + 2 > b.GetLength(0)-1))
                if ((b[x-2,y] == checkFor && b[x-1,y] == checkFor) || (b[x + 2, y] == checkFor && b[x + 1, y] == checkFor))
                    return true;

            if (!(y - 2 < 0 || y + 2 > b.GetLength(1)-1))
            {
                Debug.WriteLine($"{x}, {y}");
                if ((b[x, y - 2] == checkFor && b[x, y - 1] == checkFor) || (b[x, y + 2] == checkFor && b[x, y + 1] == checkFor))
                    return true;
            }
                

            return false;
        }

        // every row and colom should have an even number of 1's and 0's. So if the total number of 1's in a row is equal to half the with of the row a 0 should be filled in.
        private bool EvenCheck(int x, int y, int[,] b, int checkFor)
        {
            // check for a row and colum (provided that the width and height of the board is the same)
            int countRow = 0, countCol = 0;
            for (int i = 0; i < b.GetLength(0); i++)
            {
                if(b[i,y] == checkFor) countRow++;
                if(b[x,i] == checkFor) countCol++;
            }

            // check if the total number of oppisite numbers is equal to half te widht/heigt
            if (countRow == b.GetLength(0) / 2 || countCol == b.GetLength(1) / 2)
                return true;

            return false;
        }
    }
}