﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Reflection.Metadata;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace PuzzlePlayer_Namespace
{
    public enum SOLUTIONS
    {
        NONE = 0,
        UNIQUE = 1,
        MULTIPLE = 2
    }

    /*
     * This abstract class can be used to implement any kind of puzzle
     * One thing that is common for all puzzles is that a empty space is equal to -1
     * GetSolveList and IsBoardValid need to be overwriten in a subclass of Board
     */
    public abstract class Board
    {
        protected const int emptySpace = -1; // for every puzzle -1 represents a empty space
        public string description;

        // a property for getting and setting the boardsstate. The boardstate should only be setted if the imputed int[,] is valid.
        public int[,] boardState;


        // checks if the board is valid and solvable before setting the variable.
        public bool setBoardState(int[,] newState)
        {
            int[,] copy = boardState;
            boardState = newState;
            if (IsBoardValid(newState) && Solve() == SOLUTIONS.UNIQUE)
                return true;
            else
            {
                boardState = copy;
                return false;
            }
        }
        public abstract void Draw(Graphics gr, Point p, Size s);
        // a methode for solving the whole board. It uses the private SolveStep methode untill the whole board is solved
        public SOLUTIONS Solve()
        {
            // two variables for storing the result and the next solveStep
            int[,] result = boardState;
            int[,] step;

            // keep looping until the SolveStep returns null
            // if SolveStep returns null that could mean that either the whole board is solved or it is imposible to solve the board
            while ((step = SolveStep(result)) != null)
            {
                result = step;
            }

            // check if the whole board is filled
            // if not then the methode will return null because it is imposible to solve the board
            for (int i = 0; i < result.GetLength(0); i++)
                for (int j = 0; j < result.GetLength(1); j++)
                    if (result[i, j] == emptySpace)
                    {
                        return SOLUTIONS.NONE;
                    }

            boardState = result;
            return SOLUTIONS.UNIQUE;
        }

        // abstract methode for solving one step
        private int[,] SolveStep(int[,] currentBoardState)
        {
            // get a list with all the possible moves
            List<int[,]> moves = GetSolveList(currentBoardState);

            // if there are no moves found then null is returned
            if (moves.Count == 0)
                return null;

            // return one of the possible moves
            // (if the first one is always chosen than that may lead to expected behavior. For example if the possible moves are checked in a certain order)
            Random rnd = new Random();
            return moves[rnd.Next(0, moves.Count - 1)];
        }

        // a abstract methode to get a list of all possible moves
        protected abstract List<int[,]> GetSolveList(int[,] boardToSolve); 

        // abstract methode for generating a random board
        public abstract void Generate();

        // abstract methode for checking if a imputed boardstate is valid
        public abstract bool IsBoardValid(int[,] boardToCheck);
    }
}
